/*
 * Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */
#include <math.h>
#include "video.h"
#include "proto.h"
#include "dither.h"

/* Range values for lum, cr, cb. */
int LUM_RANGE;
int CR_RANGE;
int CB_RANGE;

/* Array that remaps color numbers to actual pixel values used by X server. */

unsigned char pixel[256];
PgColor_t	palette[256];

/* Arrays holding quantized value ranged for lum, cr, and cb. */

int *lum_values;
int *cr_values;
int *cb_values;

/* Declaration of global variable containing dither type. */

extern int ditherType;

/* Structures used by the X server. */

PtWidget_t	*window;

static PhImage_t *image = NULL;


/*
 *--------------------------------------------------------------
 *
 * InitColor --
 *
 *	Initialized lum, cr, and cb quantized range value arrays.
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void InitColor()
{
  int i;

  for (i=0; i<LUM_RANGE; i++) {
    lum_values[i] = ((i * 256) / (LUM_RANGE)) + (256/(LUM_RANGE*2));
  }

  for (i=0; i<CR_RANGE; i++) {
    cr_values[i] = ((i * 256) / (CR_RANGE)) + (256/(CR_RANGE*2));
  }

  for (i=0; i<CB_RANGE; i++) {
    cb_values[i] = ((i * 256) / (CB_RANGE)) + (256/(CB_RANGE*2));
  }

}


/*
 *--------------------------------------------------------------
 *
 * ConvertColor --
 *
 *	Given a l, cr, cb tuple, converts it to r,g,b.
 *
 * Results:
 *	r,g,b values returned in pointers passed as parameters.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */
static void ConvertColor(l, cr, cb, r, g, b)
     unsigned char l, cr, cb;
     unsigned char *r, *g, *b;
{
  double fl, fcr, fcb, fr, fg, fb;

  fl = (double) l;
  fcr =  ((double) cr) - 128.0;
  fcb =  ((double) cb) - 128.0;


  fr = fl + (1.40200 * fcb);
  fg = fl - (0.71414 * fcb) - (0.34414 * fcr);
  fb = fl + (1.77200 * fcr);

  if (fr < 0.0) fr = 0.0;
  else if (fr > 255.0) fr = 255.0;

  if (fg < 0.0) fg = 0.0;
  else if (fg > 255.0) fg = 255.0;

  if (fb < 0.0) fb = 0.0;
  else if (fb > 255.0) fb = 255.0;

  *r = (unsigned char) fr;
  *g = (unsigned char) fg;
  *b = (unsigned char) fb;

}

/*
 *--------------------------------------------------------------
 *
 * ResizeDisplay --
 *
 *	Resizes display window.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void ResizeDisplay(w, h)
     int w, h;
{
	PtArg_t	args[1];
	PhDim_t dim;

	dim.w = w;
	dim.h = h;
	
	PtSetArg( &args[0], Pt_ARG_DIM, &dim, 0 );
	PtSetResources( window, 1, args );
}

/*
 *--------------------------------------------------------------
 *
 * MakeWindow --
 *
 *	Create X Window
 *
 * Results:
 *	Read the code.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

static void MakeWindow( char *name) 
{
	char *title = "MPEG Play";
	PhDim_t	size;	
	PtArg_t	args[5];

	if (ditherType == NO_DITHER) return;

	PtInit( NULL );
	
	size.w = 150;
	size.h = 150;
	PtSetArg( &args[0], Pt_ARG_WINDOW_TITLE, title, 0 );
	PtSetArg( &args[1], Pt_ARG_DIM, &size, 0 );
	PtSetArg( &args[2], Pt_ARG_FILL_COLOR, Pg_TRANSPARENT, 0 );
	PtSetArg( &args[3], Pt_ARG_WINDOW_RENDER_FLAGS, 0, Ph_WM_RENDER_RESIZE );
	PtSetArg( &args[4], Pt_ARG_WINDOW_MANAGED_FLAGS, 0, Ph_WM_RESIZE );
  
	PtRealizeWidget( window = PtCreateWidget( PtWindow, NULL, 5, args ) );
}
  

/*
 *--------------------------------------------------------------
 *
 * InitDisplay --
 *
 *	Initialized display, sets up colormap, etc.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void InitDisplay( char *name )
{
  int ncolors = LUM_RANGE*CB_RANGE*CR_RANGE;
  int i, lum_num, cr_num, cb_num;
  unsigned char r, g, b;

  if (ditherType == NO_DITHER) return;

  MakeWindow(name);

  for (i=0; i<ncolors; i++) {

    lum_num = (i / (CR_RANGE*CB_RANGE))%LUM_RANGE;
    cr_num = (i / CB_RANGE)%CR_RANGE;
    cb_num = i % CB_RANGE;

    ConvertColor(lum_values[lum_num], cr_values[cr_num], cb_values[cb_num], &r, &g, &b);

	palette[i] = PgRGB( r , g, b );
//    xcolor.red = r * 256;
//    xcolor.green = g * 256;
//    xcolor.blue = b * 256;

//    if(XAllocColor(display, cmap, &xcolor) == 0 && cmap == dcmap) {
//      int j;
//      long tmp_pixel;
//      XWindowAttributes xwa;

//      if (!quietFlag) {
//	fprintf(stderr, "Using private colormap.\n");
//      }

//      /* Free colors. */
//      for(j = 0; j < i; j ++) {
//	tmp_pixel = pixel[j];
//        XFreeColors(display, cmap, &tmp_pixel, 1, 0);
//      }

//      XGetWindowAttributes(display, window, &xwa);
//      cmap = XCreateColormap(display, window, xwa.visual, AllocNone);
//      XSetWindowColormap(display, window, cmap);

//      goto retry_alloc_colors;
  //  }
    pixel[i] = i;//xcolor.pixel;
  }

  image = NULL;
}


/*
 *--------------------------------------------------------------
 *
 * InitGrayDisplay --
 *
 *	Initialized display for gray scale dither.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

#define NUM_COLORS 128

void InitGrayDisplay(name)
char *name;
{
#if 0
  int ncolors = NUM_COLORS;
  XColor xcolor;
  int i;
  Colormap dcmap;

  MakeWindow(name);

  gc = XCreateGC(display, window, 0, 0);

  dcmap = cmap = XDefaultColormap(display, DefaultScreen(display));

  xcolor.flags = DoRed | DoGreen | DoBlue;

  retry_alloc_grays:
  for (i=0; i<ncolors; i++) {

    xcolor.red = (i*2) * 256;
    xcolor.green = (i*2) * 256;
    xcolor.blue = (i*2) * 256;

    if(XAllocColor(display, cmap, &xcolor) == 0 && cmap == dcmap) {
      int j;
      long tmp_pixel;
      XWindowAttributes xwa;

      if (!quietFlag) {
	fprintf(stderr, "Using private colormap.\n");
      }

      /* Free colors. */
      for(j = 0; j < i; j ++) {
	tmp_pixel = pixel[j*2];
        XFreeColors(display, cmap, &tmp_pixel, 1, 0);
      }

      XGetWindowAttributes(display, window, &xwa);
      cmap = XCreateColormap(display, window, xwa.visual, AllocNone);
      XSetWindowColormap(display, window, cmap);

      goto retry_alloc_grays;
    }
    pixel[(i*2)] = xcolor.pixel;
    pixel[(i*2)+1] = xcolor.pixel;
  }

  ximage = NULL;
#endif
}


/*
 *--------------------------------------------------------------
 *
 * InitMonoDisplay --
 *
 *	Initialized display for monochrome dither.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void InitMonoDisplay( char *name )
{
#if 0
  XGCValues xgcv;

  MakeWindow(name);

  xgcv.background = BlackPixel(display, DefaultScreen(display));
  xgcv.foreground = WhitePixel(display, DefaultScreen(display));

  gc = XCreateGC(display, window, GCForeground | GCBackground, &xgcv);

  ximage = NULL;
#endif
}


/*
 *--------------------------------------------------------------
 *
 * InitColorDisplay --
 *
 *	Initialized display for full color output.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void InitColorDisplay( char *name )
{

	MakeWindow(name);
	
	image = NULL;
}


/*
 *--------------------------------------------------------------
 *
 * ExecuteDisplay --
 *
 *	Actually displays display plane in previously created window.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

void ExecuteDisplay( VidStream *vid_stream )
{

	totNumFrames++;
	if (!quietFlag) {
		fprintf (stderr, "%d\r", totNumFrames);
	}

#if 0
	  if (ximage == NULL) {
		if (ditherType == Twox2_DITHER) {
			ximage = XCreateImage(display, None, 8, ZPixmap, 0, &dummy,
			    vid_stream->mb_width * 32,
			    vid_stream->mb_height * 32, 8, 0);
		} else if (ditherType == FULL_COLOR_DITHER) {
			fc_visual = FindFullColorVisual(display, &depth);
			ximage = XCreateImage (display, fc_visual, depth, ZPixmap,
			     0, &dummy, vid_stream->mb_width * 16,
			     vid_stream->mb_height * 16, 32, 0);
		} else if (ditherType == MONO_DITHER || ditherType == MONO_THRESHOLD) {
			ximage = XCreateImage (display, None, 1, XYBitmap, 0, &dummy,
			     vid_stream->mb_width * 16,
			     vid_stream->mb_height * 16, 8, 0);
			ximage->byte_order = MSBFirst;
			ximage->bitmap_bit_order = MSBFirst;
		} else {
			ximage = XCreateImage(display, None, 8, ZPixmap, 0, &dummy,
			    vid_stream->mb_width * 16,
			    vid_stream->mb_height * 16, 8, 0);
		}
	}
#endif
	if (!noDisplayFlag) {
		PhPoint_t pos;
		pos.x = pos.y = 0;
		PgSetRegion( window->rid );
		if ( ditherType != FULL_COLOR_DITHER) {
			PgSetPalette( palette, 0, 0, 256, Pg_PALSET_SOFT, 0 );
			if ( ditherType == Twox2_DITHER ) {
				PhPoint_t	size;
				size.x = vid_stream->current->image->size.w * 2;
				size.y = vid_stream->current->image->size.h * 2;
				PgDrawImagemx( vid_stream->current->image->image, Pg_IMAGE_PALETTE_BYTE, &pos, &size,
						 vid_stream->current->image->size.w * 2, 0 );
			} else
				PgDrawImagemx( vid_stream->current->image->image, Pg_IMAGE_PALETTE_BYTE, &pos, 
						 (PhPoint_t *)&vid_stream->current->image->size, 
						 vid_stream->current->image->size.w, 0 );
		} else
			PgDrawImagemx( vid_stream->current->image->image, Pg_IMAGE_DIRECT_8888, &pos, 
						 (PhPoint_t *)&vid_stream->current->image->size, 
						 vid_stream->current->image->size.w * 4, 0 );
		PgFlush();
	}
}


